/**************************************************************************\
 *  aRT : API R - TerraLib                                                *
 *  Copyright (C) 2003-2005  LEG                                          *
 *                                                                        *
 *  This program is free software; you can redistribute it and/or modify  *
 *  it under the terms of the GNU General Public License as published by  *
 *  the Free Software Foundation; either version 2 of the License, or     *
 *  (at your option) any later version.                                   *
 *                                                                        *
 *  This program is distributed in the hope that it will be useful,       *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *  GNU General Public License for more details.                          *
 *                                                                        *
 *  You should have received a copy of the GNU Lesser General Public      *
 *  License along with this library.                                      *
\**************************************************************************/

#include "aRTquerier.h"
#include "SEXPutils.h"

#include <TeQuerierParams.h>
#include <TeQuerier.h>
#include <TeInitQuerierStrategy.h>

using namespace std;

aRTquerier::aRTquerier(TeQuerierParams& params)
{
	TeInitQuerierStrategies(); // never forget it!

	Querier = new TeQuerier(params);
	if(! Querier->getNumTimeFrames() )
	{
		Querier->loadInstances();
	}

	cout << "slides: " << Querier->getNumTimeFrames() << endl;
	cout << "instances: " << Querier->numElemInstances () << endl;
	ActualSlide = -1;
}

SEXP aRTquerier::GetData()
{
	TeSTInstance sti;
	TeMultiGeometry mg;

    if( !Querier->fetchInstance(sti) )
	{
		warning("Could not fetch the instance\n");
		return R_NilValue;
	}

	TeProperty prop;

	SEXP colnames, result, rownames, list;
	SEXP *tempdata;
	vector<string> tempnames;

	tempdata = new SEXP[4];  // 3 geom + 1 att
	int numdata = 0; // number of data items stored in the instance
	                 // and each result found is stored in temp
	int unprotect_ = 0;
	
	// process attributes
	unsigned size = sti.getPropertyVector().size();
	if(size)
	{
    	list     = PROTECT(allocVector( VECSXP, size ));
    	rownames = PROTECT(allocVector( STRSXP, 1    ));
    	colnames = PROTECT(allocVector( STRSXP, size ));
		unprotect_ +=3;

		SEXP data;

		SET_STRING_ELT(rownames, 0, mkChar("1"));
		
		for(unsigned i = 0; i < size; i++)
		{
			sti.getProperty(prop, i);
			PROTECT(data = allocVector(STRSXP, 1));
			unprotect_++;
			
			SET_STRING_ELT(data,     0, mkChar(prop.value_.c_str()));
			SET_STRING_ELT(colnames, i, mkChar(prop.attr_.rep_.name_.c_str()));
			SET_VECTOR_ELT(list,     i, data);
		}
		setAttrib(list, R_NamesSymbol, colnames);
		PROTECT(tempdata[numdata] = AsDataFrame(list, rownames));
		unprotect_++;
		numdata++;
		tempnames.push_back("data");
	}
	
	// process geometries
	mg = sti.geometries();

	if( mg.hasPolygons() )
	{
		unprotect_++;
		PROTECT(tempdata[numdata] = GetPolygonsTeMultiGeometry(mg));
		tempnames.push_back(aRTstrPolygon);
		numdata++;
	}
	if( mg.hasLines() )
	{
		unprotect_++;
		PROTECT(tempdata[numdata] = GetLinesTeMultiGeometry(mg));
		tempnames.push_back(aRTstrLine);
		numdata++;
	}
	if( mg.hasPoints() )
	{
		unprotect_++;
		PROTECT(tempdata[numdata] = GetPointsTeMultiGeometry(mg));
		tempnames.push_back(aRTstrPoint);
		numdata++;
	}

	if( !numdata ) return R_NilValue;
   	result   = PROTECT(allocVector( VECSXP, numdata ));
    colnames = PROTECT(allocVector( STRSXP, numdata ));
	unprotect_+=2;

	// copy the results pointers
	for(int i = 0; i < numdata; i++)
	{
		SET_VECTOR_ELT(result, i, tempdata[i]);
		SET_STRING_ELT(colnames, i, mkChar(tempnames[i].c_str()));
	}

	setAttrib(result, R_NamesSymbol, colnames);
	delete[] tempdata;
	
	UNPROTECT(unprotect_);
	return result;
}

aRTquerier::~aRTquerier()
{
	PrintSilent("Deleting one aRTquerier object from memory\n");
	delete Querier;
}
	
// this is the order to appear inside R. S is from Summary
enum aRTquerierSenum
{
	aRTquerierSClass = 0,
	aRTquerierSElements,
	aRTquerierSSlides,
	aRTquerierSCurrent,
	aRTquerierSSize
};
					   
SEXP aRTquerier::Summary()
{
    SEXP info, names;

	info    = allocVector(VECSXP, aRTquerierSSize);  // List containing the connection attributes
    names   = allocVector(STRSXP, aRTquerierSSize);  // Names Vector

	STRING_LIST_ITEM(info,  names, aRTquerierSClass,    "class",    Class());
    INTEGER_LIST_ITEM(info, names, aRTquerierSElements, "elements", Querier->numElemInstances());
    INTEGER_LIST_ITEM(info, names, aRTquerierSSlides,   "slides",   Querier->getNumTimeFrames());
    INTEGER_LIST_ITEM(info, names, aRTquerierSCurrent,  "current",  ActualSlide);
	
	setAttrib(info, R_NamesSymbol, names); // Set the names of the list

    return info;
}

const string aRTquerier::Print()
{
	stringstream s;

	if (!IsValid()) error("Invalid object!");
		 
	 s << artOBJECT("aRTquerier")   << "\n\n"
	   << "Slides: "       << Querier->getNumTimeFrames()  << endl
	   << "Current slide: " << ActualSlide                  << endl
       << "Elements of current slide: " << Querier->numElemInstances()  << endl;
/*
	TeTemporalSeries ts;
	if( Querier->getTS (ts) )
	{
		s << "Temporal Series:" << "\n\n"
	      << "";
	}
	else
		s << "No temporal serie" << endl;
*/
	return s.str();
}

SEXP aRTquerier::NextSlide()
{
	if(ActualSlide >= Querier->getNumTimeFrames() - 1) // it begins from -1
		error("Next slide does not exist\n");
	
	ActualSlide++;
	Querier->loadInstances(ActualSlide);
	
	SEXP resp;
	PROTECT(resp = allocVector(INTSXP, 1));
	INTEGER(resp)[0] = Querier->numElemInstances ();
	UNPROTECT(1);
	return resp;
}

bool aRTquerier::IsValid()
{
	if(Valid && !Querier)
	{
		Valid = false;
	}

	return Valid;
}

